@@ -3,6 +3,8 @@ import logging
 import os
 import sys
 
+from typing import List
+from itertools import product
 import numpy as np
 import pytest
 
@@ -82,6 +84,17 @@ class TestPipeline:
         count2 = r.execute_streaming(chunk_size=100)
         assert count == count2
 
+
+    @pytest.mark.parametrize("filename", ["range.json", "range.py"])
+    def test_subsetstreaming(self, filename):
+        """Can we fetch a subset of PDAL dimensions as a numpy array while streaming"""
+        r = get_pipeline(filename)
+        limit = ['X','Y','Z','Intensity']
+        arrays = list(r.iterator(chunk_size=100,allowed_dims=limit))
+        assert len(arrays) == 11
+        assert len(arrays[0].dtype) == 4
+
+
     @pytest.mark.parametrize("filename", ["sort.json", "sort.py"])
     def test_execute_streaming_non_streamable(self, filename):
         r = get_pipeline(filename)
@@ -113,6 +126,18 @@ class TestPipeline:
         assert a[0][0] == 635619.85
         assert a[1064][2] == 456.92
 
+    @pytest.mark.parametrize("filename", ["sort.json", "sort.py"])
+    def test_subsetarray(self, filename):
+        """Can we fetch a subset of PDAL dimensions as a numpy array"""
+        r = get_pipeline(filename)
+        limit = ['X','Y','Z']
+        r.execute(allowed_dims=limit)
+        arrays = r.arrays
+        assert len(arrays) == 1
+        assert len(arrays[0].dtype) == 3
+
+
+
     @pytest.mark.parametrize("filename", ["sort.json", "sort.py"])
     def test_metadata(self, filename):
         """Can we fetch PDAL metadata"""
@@ -140,8 +165,11 @@ class TestPipeline:
         """Can we fetch PDAL pipeline string"""
         r = get_pipeline(filename)
         r.execute()
-        assert json.loads(r.pipeline) == {
-            "pipeline": [
+        # filename might be an object in PDAL 2.9+
+        # https://github.com/PDAL/PDAL/issues/4751
+
+        returned = json.loads(r.pipeline)
+        expected = { "pipeline": [
                 {
                     "filename": "test/data/1.2-with-color.las",
                     "tag": "readers_las1",
@@ -155,6 +183,10 @@ class TestPipeline:
                 },
             ]
         }
+        try:
+            assert returned['pipeline'][0]['filename'] == "test/data/1.2-with-color.las"
+        except AttributeError:
+            assert returned['pipeline'][0]['filename']['path'] == "test/data/1.2-with-color.las"
 
     @pytest.mark.parametrize("filename", ["sort.json", "sort.py"])
     def test_no_execute(self, filename):
@@ -371,6 +403,15 @@ class TestPipeline:
         assert 'readers.las' in info.keys()
         assert info['readers.las']['num_points'] == 1065
 
+    def test_quickinfo_offsets_scales(self):
+        r = pdal.Reader(os.path.join(DATADIRECTORY,"simple.laz"))
+        p = r.pipeline()
+        info = p.quickinfo
+        assert 'readers.las' in info.keys()
+        assert 'offset_x' in info['readers.las']['metadata'].keys()
+        assert 'scale_x' in info['readers.las']['metadata'].keys()
+        assert info['readers.las']['num_points'] == 1065
+
     def test_jsonkwarg(self):
         pipeline = pdal.Reader(os.path.join(DATADIRECTORY,"autzen-utm.las")).pipeline().toJSON()
         r = pdal.Pipeline(json=pipeline)
@@ -541,6 +582,65 @@ class TestDataFrame:
             assert data["Intensity"].sum() == 57684
 
 
+class TestGeoDataFrame:
+
+    @pytest.mark.skipif(
+        not pdal.pipeline.GeoDataFrame,
+        reason="geopandas is not available",
+    )
+    def test_fetch(self):
+        r = pdal.Reader(os.path.join(DATADIRECTORY,"autzen-utm.las"))
+        p = r.pipeline()
+        p.execute()
+        record_count = p.arrays[0].shape[0]
+        dimension_count = len(p.arrays[0].dtype)
+        gdf = p.get_geodataframe(0)
+        gdf_xyz = p.get_geodataframe(0, xyz=True)
+        gdf_crs = p.get_geodataframe(0, crs="EPSG:4326")
+        assert len(gdf) == record_count
+        assert len(gdf.columns) == dimension_count + 1
+        assert isinstance(gdf, pdal.pipeline.GeoDataFrame)
+        assert gdf.geometry.is_valid.all()
+        assert not gdf.geometry.is_empty.any()
+        assert gdf.crs is None
+        assert gdf.geometry.z.isna().all()
+        assert not gdf_xyz.geometry.z.isna().any()
+        assert gdf_crs.crs.srs == "EPSG:4326"
+
+    @pytest.mark.skipif(
+        not pdal.pipeline.GeoDataFrame,
+        reason="geopandas is not available",
+    )
+    def test_load(self):
+        r = pdal.Reader(os.path.join(DATADIRECTORY,"autzen-utm.las"))
+        p = r.pipeline()
+        p.execute()
+        data = p.arrays[0]
+        gdf = pdal.pipeline.GeoDataFrame(
+            data,
+            geometry=pdal.pipeline.points_from_xy(data["X"], data["Y"], data["Z"])
+        )
+        dataframes = [gdf, gdf, gdf]
+        filter_intensity = """{
+          "pipeline":[
+            {
+              "type":"filters.range",
+              "limits":"Intensity[100:300)"
+            }
+          ]
+        }"""
+        p = pdal.Pipeline(filter_intensity, dataframes = dataframes)
+        p.execute()
+        arrays = p.arrays
+        assert len(arrays) == 3
+
+        # We copied the array three times. Sum the Intensity values
+        # post filtering to see if we had our intended effect
+        for data in arrays:
+            assert len(data) == 387
+            assert data["Intensity"].sum() == 57684
+
+
 class TestPipelineIterator:
     @pytest.mark.parametrize("filename", ["sort.json", "sort.py"])
     def test_non_streamable(self, filename):
@@ -646,3 +746,149 @@ class TestPipelineIterator:
             np.testing.assert_array_equal(a1, a2)
         assert next(it1, None) is None
         assert next(it2, None) is None
+
+
+def gen_chunk(count, random_seed = 12345):
+    rng = np.random.RandomState(count*random_seed)
+    # Generate dummy data
+    result = np.zeros(count, dtype=[("X", float), ("Y", float), ("Z", float)])
+    result['X'][:] = rng.uniform(-2, -1, count)
+    result['Y'][:] = rng.uniform(1, 2, count)
+    result['Z'][:] = rng.uniform(3, 4, count)
+    return result
+
+
+class TestPipelineInputStreams():
+
+    # Test cases
+    ONE_ARRAY_FULL = [[gen_chunk(1234)]]
+    MULTI_ARRAYS_FULL = [*ONE_ARRAY_FULL, [gen_chunk(4321)]]
+
+    ONE_ARRAY_STREAMED = [[gen_chunk(10), gen_chunk(7), gen_chunk(3), gen_chunk(5), gen_chunk(1)]]
+    MULTI_ARRAYS_STREAMED = [*ONE_ARRAY_STREAMED, [gen_chunk(5), gen_chunk(2), gen_chunk(3), gen_chunk(1)]]
+
+    MULTI_ARRAYS_MIXED = [
+        *MULTI_ARRAYS_STREAMED,
+        *MULTI_ARRAYS_FULL
+    ]
+
+    @pytest.mark.parametrize("in_arrays_chunks, use_setter", [
+        (arrays_chunks, use_setter) for arrays_chunks, use_setter in product([
+            ONE_ARRAY_FULL, MULTI_ARRAYS_FULL,
+            ONE_ARRAY_STREAMED, MULTI_ARRAYS_STREAMED,
+            MULTI_ARRAYS_MIXED
+        ], ['False', 'True'])
+    ])
+    def test_pipeline_run(self, in_arrays_chunks, use_setter):
+        """
+        Test case to validate possible usages:
+        - Combining "full" arrays and "streamed" ones
+        - Setting input arrays through the Pipeline constructor or the setter
+        """
+        # Assuming stream mode for lists that contain more than one chunk.
+        # And that first chunk is the biggest of all, to simplify input buffer size creation.
+        in_arrays = [
+            np.zeros(chunks[0].shape, chunks[0].dtype) if len(chunks) > 1 else chunks[0]
+            for chunks in in_arrays_chunks
+        ]
+
+        def get_stream_handler(in_array, in_array_chunks):
+            in_array_chunks_it = iter(in_array_chunks)
+            def load_next_chunk():
+                try:
+                    next_chunk = next(in_array_chunks_it)
+                except StopIteration:
+                    return 0
+
+                chunk_size = next_chunk.size
+                in_array[:chunk_size]["X"] = next_chunk[:]["X"]
+                in_array[:chunk_size]["Y"] = next_chunk[:]["Y"]
+                in_array[:chunk_size]["Z"] = next_chunk[:]["Z"]
+
+                return chunk_size
+
+            return load_next_chunk
+
+        stream_handlers = [
+            get_stream_handler(arr, chunks) if len(chunks) > 1 else None
+            for arr, chunks in zip(in_arrays, in_arrays_chunks)
+        ]
+
+        expected_count = sum([sum([len(c) for c in chunks]) for chunks in in_arrays_chunks])
+
+        pipeline = """
+        {
+            "pipeline": [{
+                "type": "filters.stats"
+            }]
+        }
+        """
+        if use_setter:
+            p = pdal.Pipeline(pipeline)
+            p.inputs = [(a, h) for a, h in zip(in_arrays, stream_handlers)]
+        else:
+            p = pdal.Pipeline(pipeline, arrays=in_arrays, stream_handlers=stream_handlers)
+
+        count = p.execute()
+        out_arrays = p.arrays
+        assert count == expected_count
+        assert len(out_arrays) == len(in_arrays)
+
+        for in_array_chunks, out_array in zip(in_arrays_chunks, out_arrays):
+            np.testing.assert_array_equal(out_array, np.concatenate(in_array_chunks))
+
+    @pytest.mark.parametrize("in_arrays, use_setter", [
+        (arrays, use_setter) for arrays, use_setter in product([
+            [c[0] for c in ONE_ARRAY_FULL],
+            [c[0] for c in MULTI_ARRAYS_FULL]
+        ], ['False', 'True'])
+    ])
+    def test_pipeline_run_backward_compat(self, in_arrays, use_setter: bool):
+        expected_count = sum([len(a) for a in in_arrays])
+
+        pipeline = """
+        {
+            "pipeline": [{
+                "type": "filters.stats"
+            }]
+        }
+        """
+        if use_setter:
+            p = pdal.Pipeline(pipeline)
+            p.inputs = in_arrays
+        else:
+            p = pdal.Pipeline(pipeline, arrays=in_arrays)
+
+        count = p.execute()
+        out_arrays = p.arrays
+        assert count == expected_count
+        assert len(out_arrays) == len(in_arrays)
+
+        for in_array, out_array in zip(in_arrays, out_arrays):
+            np.testing.assert_array_equal(out_array, in_array)
+
+    @pytest.mark.parametrize("in_array, invalid_chunk_size", [
+        (in_array, invalid_chunk_size) for in_array, invalid_chunk_size in product(
+            [gen_chunk(1234)],
+            [-1, 12345])
+    ])
+    def test_pipeline_fail_with_invalid_chunk_size(self, in_array, invalid_chunk_size):
+        """
+        Ensure execution fails when using an invalid stream handler:
+        - One that returns a negative chunk size
+        - One that returns a chunk size bigger than the buffer capacity
+        """
+        was_called = False
+        def invalid_stream_handler():
+            nonlocal was_called
+            if was_called:
+                # avoid infinite loop
+                raise ValueError("Invalid handler should not have been called a second time")
+            was_called = True
+            return invalid_chunk_size
+
+        p = pdal.Pipeline(arrays=[in_array], stream_handlers=[invalid_stream_handler])
+        with pytest.raises(RuntimeError,
+                           match=f"Stream chunk size not in the range of array length: {invalid_chunk_size}"):
+            p.execute()
+
